#!/usr/bin/perl -w
#download NCBI genome porject information page as a text file
#using eutil calls
#store the result as a .gpt (genome project text) file
#Author: William Hsiao
#May 2006

use strict;
use LWP::Simple;
use XML::Simple;
use Getopt::Long;
use Data::Dumper;
use FileHandle;

my $directory = $ARGV[0] || "/home/shared/NCBI_genomes/curated/Bacteria";


my $utils     = "http://www.ncbi.nlm.nih.gov/entrez/eutils";
my $report    = 'Summary';
my $database  = 'Genome';
my $logfile   = 'log/NCBI_EUTIL.log';
my $debug     = 0;

GetOptions(
    "dir=s"      => \$directory,
    "Report=s"   => \$report,
    "Database=s" => \$database,
    "debug"      => \$debug,
    "logfile=s"  => \$logfile
);

my ( $day, $month, $year ) = (localtime)[ 3, 4, 5 ];
$month = $month + 1;
$year  = $year + 1900;

open( LOG, ">>$logfile" ) or warn "Unable to open Logfile $logfile\n";
LOG->autoflush(1);
print LOG "Starting eutil_download2.pl on $year-$month-$day:\n";
my $dir_filename_ref = &scan_dir($directory);
foreach ( keys %$dir_filename_ref ) {
    my $dirname = $_;
    foreach ( @{ $dir_filename_ref->{$dirname} } ) {
        my $accnum = $_;
        
        #do not overwrite existing gpt files (may comment out)
        next if(-e "$directory/$dirname/$accnum.gpt" );
        chomp $accnum;
        my $gpid = &accession2gpid($accnum);
        if ( $gpid =~ /^Bad Request/ ) {
            print LOG $gpid;
            next;
        }
        else {
            print LOG "Found GPID $gpid for ACCNUM $accnum\n";
        }

     #had to put in a workaround due to problems with getting the correct gpid
     #Note: I am trying both ids below, however, after running it looks like the first id will always work!
        my $gptxt;
        if ( ref($gpid) ne "ARRAY" ) {
            $gptxt = &get_genomeprj_txt($gpid);
            if ( $gptxt =~ /Bad Request/ ) {
                print LOG $gptxt;
                next;
            }
            else {
                print LOG "Found Text for GPID $gpid\n";
            }
        }
        else {
            $gptxt = &get_genomeprj_txt( $gpid->[0] );
            if ( $gptxt =~ /Bad Request/ ) {
                print LOG $gptxt;
                $gptxt = &get_genomeprj_txt( $gpid->[1] );
                if ( $gptxt =~ /Bad Request/ ) {
                    print LOG $gptxt;
                    next;
                }    
            }
            else {
                print LOG "Found Text for GPID $gpid\n";
            }
        }
        open( GPTWRITE, ">$directory/$dirname/$accnum.gpt" )
          or die "Can't open output gpt file in $directory/$dirname/\n";
        print GPTWRITE $gptxt;
        close GPTWRITE;
        sleep 5;
    }
}
print LOG "End eutil_download2.pl on $year-$month-$day:\n";
close LOG;

sub scan_dir {
    my $parentdir = shift;
    my %dirname_filenames
      ; #key is the subdir name and value is an array of file names in that dir
    opendir( DIR, $parentdir ) or die "cannot opendir $parentdir: $!";
    while ( defined( my $sdir = readdir(DIR) ) ) {
        my $filenotfound = 0;
        next unless ( -d "$parentdir/$sdir" );
        next if ( $sdir eq '.' );
        next if ( $sdir eq '..' );
        opendir( SUBDIR, "$parentdir/$sdir" )
          or die "cannot opendir $sdir: $!";
        while ( defined( my $file = readdir(SUBDIR) ) ) {
            if ( $file =~ /(NC_\d+)(.gbk)$/ ) {
                push @{ $dirname_filenames{$sdir} }, $1;
            }
        }
    }
    return \%dirname_filenames;
}

sub get_genomeprj_txt {
    my $gpid            = shift;
    my $genomeprj_fetch =
      'http://eutils.ncbi.nlm.nih.gov/entrez/utils/pmfetch.cgi?db=genomeprj&report=Overview&mode=asn&id='
      . $gpid;
    my $genomeprj_txt = get($genomeprj_fetch);
    return "Bad Request: Unable to get the genome project txt for $gpid\n"
      unless ( defined $genomeprj_txt );
    if ($debug) { print "GENOME PROJECT RESULT:\n$genomeprj_txt\n\n"; }
    return $genomeprj_txt;
}

sub accession2gpid {
    my $accession = shift;
    my $esearch   =
      "$utils/esearch.fcgi?db=$database&retmax=1&usehistory=y&term=";
    my $esearch_result = get( $esearch . $accession );
    return "Bad Request: Unable to get the esearch result for $accession\n"
      unless ( defined $esearch_result );
    $esearch_result =~
      m|<Count>(\d+)</Count>.*<QueryKey>(\d+)</QueryKey>.*<WebEnv>(\S+)</WebEnv>|s;
    my $Count            = $1;
    my $QueryKey         = $2;
    my $WebEnv           = $3;
    my $genome2genomeprj =
      'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?db=genomeprj&dbfrom=genome&&WebEnv='
      . $WebEnv
      . '&query_key='
      . $QueryKey;
    my $projectdata = get($genome2genomeprj);
    return
      "Bad Request: Unable to link to genome project from genome for $accession\n"
      unless ( defined $projectdata );
    my $xs     = new XML::Simple;
    my $xmlref = $xs->XMLin($projectdata);

#Sometimes there are two ids in the XML and only one is the right one (and there is no way to tell which one is right!)
    my $projectid;
    if ( ref( $xmlref->{LinkSet}->{LinkSetDb}{Link} ) eq "HASH" ) {
        $projectid = $xmlref->{LinkSet}->{LinkSetDb}{Link}{Id};
    }
    else {
        $projectid = [
            $xmlref->{LinkSet}->{LinkSetDb}{Link}[0]{Id},
            $xmlref->{LinkSet}->{LinkSetDb}{Link}[1]{Id}
        ];
    }

    return
      "Bad Request: Unable to find projectid in the XML file for $accession\n"
      unless ( defined $projectid );
    return $projectid;
}

sub trim {
    my $string = shift;
    $string =~ s/\n+//g;
    $string =~ s/^\s+//g;
    $string =~ s/\s+$//g;
    return $string;
}
