#!/usr/bin/perl
#This runs cvtree using the wrapper Cvtree.pm

use warnings;
use strict;

use lib "../perl_lib";
use lib "../microbedb/MicrobeDB";
use Cvtree;
use File::Temp qw/ tempfile :mktemp/;

#use the microbedb api
use Search;
use Buster;

my $tmp_dir_root      = '~/tmp/';
my $microbedb_version = 1;

my $parallel = 0;

my $rep_obj =
  new Replicon( version_id => $microbedb_version, rep_type => 'chromosome' );

#Get all chromosomes
my $search_obj  = new Search();
my @chromosomes = $search_obj->object_search($rep_obj);

my $dsn = "DBI:mysql:database=microbedb;host=localhost";
my $dbh = DBI->connect( $dsn, 'perlapi', 'microbedb', { RaiseError => 1 } )
  || die $DBI::errstr;

#Get a list of distances that have already been calculated
my $precomputed = load_precomputed();

#Check nodes on buster
my $buster_obj = new Buster();

my @jobs;
my $total_count = 0;
for ( my $i = 0; $i < scalar(@chromosomes); $i++ ) {
    my $id = $chromosomes[$i]->rep_accnum();
    print "id1:$i\t$id\n";

    for ( my $j = $i + 1; $j < scalar(@chromosomes); $j++ ) {
        $total_count++;
        my $rep_accnum1 = $chromosomes[$i]->rep_accnum();
        my $rep_accnum2 = $chromosomes[$j]->rep_accnum();
        unless (
            dist_already_calculated(
                $chromosomes[$i]->rep_accnum(),
                $chromosomes[$j]->rep_accnum()
            )
          )
        {
            push( @jobs, [ $chromosomes[$i], $chromosomes[$j] ] );
        }

        #group jobs together unless we are at the very last iteration
        if ( scalar(@jobs) >= 20 ) {
            run_jobs(@jobs);
            @jobs = ();
        }
    }
}

#Run any jobs left over on the queue
my $not_run = scalar(@jobs);
if ( $not_run > 0 ) {
    run_jobs(@jobs);
}

print "total_distances: $total_count \n";

sub run_jobs {
    my @run_jobs = @_;

    if ($parallel) {
        my $node      = $buster_obj->get_node();
        my $job_count = scalar(@run_jobs);
        print "Sending $job_count jobs to node: $node\n";
        if ( my $pid = fork() ) {

            #       if (1) {
            my $tmp_dir = $tmp_dir_root . mktemp("cvdirXXXXXX");
            mkdir($tmp_dir);
            my $cvtree_obj = new Cvtree(
                buster  => 1,
                node    => $node,
                tmp_dir => $tmp_dir,
            );
            foreach my $curr_job (@run_jobs) {

                #get the filenames for the faa files
                my $seq1 = $curr_job->[0]->get_filename('faa');
                my $seq2 = $curr_job->[1]->get_filename('faa');

                #calculate the distance between the chromosomes using cvtree
                my $result = $cvtree_obj->run( $seq1, $seq2 );
                insert_dist( $curr_job->[0]->rep_accnum(),
                    $curr_job->[1]->rep_accnum(), $result );
            }
            `rm -rf $tmp_dir`;

            exit;

        } else {

     #child of the fork
     #wait a couple of seconds before moving to the next iteration in the loop
            sleep(1);
        }
    } else {

        my $job_count = scalar(@run_jobs);
        print "Sending $job_count jobs";

        #       if (1) {
        my $tmp_dir = $tmp_dir_root . mktemp("cvdirXXXXXX");
        mkdir($tmp_dir);
        my $cvtree_obj = new Cvtree(
            buster  => 0,
            tmp_dir => $tmp_dir,
        );
        foreach my $curr_job (@run_jobs) {

            #get the filenames for the faa files
            my $seq1 = $curr_job->[0]->get_filename('faa');
            my $seq2 = $curr_job->[1]->get_filename('faa');

            #calculate the distance between the chromosomes using cvtree
            my $result = $cvtree_obj->run( $seq1, $seq2 );
            insert_dist( $curr_job->[0]->rep_accnum(),
                $curr_job->[1]->rep_accnum(), $result );
        }
        `rm -rf $tmp_dir`;    

    }

}

#Inserts the distance into the database
sub insert_dist {
    my @values = @_;
    my @fields = qw(rep_accnum1 rep_accnum2 distance);

    my $dsn = "DBI:mysql:database=microbedb;host=localhost";
    my $dbh = DBI->connect( $dsn, 'perlapi', 'microbedb', { RaiseError => 1 } )
      || die $DBI::errstr;

#Need to add back ticks around each field name because certain fields (eg. order) will cause problems
    my @new_fields;
    foreach (@fields) {
        push( @new_fields, "`$_`" );
    }

    #Make the array of fields into a comma seperated string
    my $fields = join( ',', @new_fields );

 #Makes a string of ?'s so we can bind to them later
 #Note: always bind the values because it looks after converting undef to null
    my $bind = join( ',', ('?') x @values );

    #my $sql = "INSERT IGNORE $table_name ($fields) VALUES ($bind)";

#Use REPLACE instead of INSERT so this method can be used for both inserts or replacement
    my $sql = "REPLACE replicon_distance ($fields) VALUES ($bind)";

    #Create new genomeproject recordicrobedb
    my $sth = $dbh->prepare($sql);

    #call the statement
    $sth->execute(@values);

}

sub load_precomputed {

    my $sql = " Select * from replicon_distance";

    #Create new genomeproject recordicrobedb
    my $sth = $dbh->prepare($sql);

    #call the statement
    $sth->execute();
    my $precomputed;
    while ( my $result = $sth->fetchrow_hashref ) {
        $precomputed->{ $result->{rep_accnum1} }{ $result->{rep_accnum2} } =
          1;
    }
    return $precomputed;
}

sub dist_already_calculated {
    my ( $rpvid1, $rpvid2 ) = @_;

    if (   $precomputed->{$rpvid1}{$rpvid2}
        || $precomputed->{$rpvid2}{$rpvid1} )
    {
        return 1;
    } else {
        return 0;
    }

}

