
package Genome;
#
#Contains information on sequenced bacterial genomes
#For documentation use: perldoc Genome.pm

use strict;
use warnings;
our $AUTOLOAD; # before Perl 5.6.0 say "use vars '$AUTOLOAD';"
use Carp;

# Class data and methods, that refer to the collection of all objects
# in the class, not just one specific object
{
    my $_count = 0;
    sub get_count {
        $_count;
    }
    sub _incr_count {
        ++$_count;
    }
    sub _decr_count {
        --$_count;
    }
}

# Parses the data file once
# This part will vary depending on the source of the genome data
my %_genome_desc;

my $desc_file = "/home/mlangill/research/db/Org_db_03Mar06.csv";

#Creates the data structure from the tab delimited file

open(DESC,$desc_file) or die "Can't open $desc_file for reading: $!";

<DESC>;

foreach(<DESC>){
    chomp;
    my @desc =split(/\t/);
    $_genome_desc{$desc[22]}{'_tax_id'}=$desc[1];
    $_genome_desc{$desc[22]}{'_organism_name'}=$desc[2];
    $_genome_desc{$desc[22]}{'_organism_dir'}=$desc[3];
    $_genome_desc{$desc[22]}{'_list_accessions'}=$desc[4];
    $_genome_desc{$desc[22]}{'_gram_stain'}=$desc[5];
    $_genome_desc{$desc[22]}{'_num_chromosomes'}=$desc[6];
    $_genome_desc{$desc[22]}{'_num_plasmids'}=$desc[7];
    $_genome_desc{$desc[22]}{'_size'}=$desc[25];
    $_genome_desc{$desc[22]}{'_gc_content'}=$desc[19];
    $_genome_desc{$desc[22]}{'_accession'}=$desc[22];
    $_genome_desc{$desc[22]}{'_gi'}=$desc[23];
    $_genome_desc{$desc[22]}{'_replicon_type'}=$desc[24];
    $_genome_desc{$desc[22]}{'_num_genes'}=$desc[29];
    $_genome_desc{$desc[22]}{'_phylum'}=$desc[33];
    $_genome_desc{$desc[22]}{'_class'}=$desc[34];
    $_genome_desc{$desc[22]}{'_order'}=$desc[35];
    $_genome_desc{$desc[22]}{'_family'}=$desc[36];
    $_genome_desc{$desc[22]}{'_genus'}=$desc[37];
    $_genome_desc{$desc[22]}{'_species'}=$desc[38];
    $_genome_desc{$desc[22]}{'_subspecies'}=$desc[39]; 
    
}

close(DESC);

my $dist_matrix_file="/home/mlangill/research/db/dist_matrix.txt";

open(MATRIX,$dist_matrix_file) || croak "Can't open $dist_matrix_file: $!";


#first line in file contains the number of genomes in the file or matrix
my $num_genomes=<MATRIX>;
chomp($num_genomes);


my (@genome_names,%usable_genomes, @matrix);
while(<MATRIX>){
    chomp;
    #split the line at each space
    my @matrix_line = split(/\s+/);

    #take the genome id from the first cell and put it on the genome name array
    push(@genome_names,shift(@matrix_line));

    #put the matrix line onto the 2d array matrix
    push(@matrix,[@matrix_line]);

}

close(MATRIX);

#put the data into a 2d hash
#Note: Each distance will be put in the hash twice so that the order of the keys used to access the distance does not matter
my %distance_hash;
for(my $i=1; $i < $num_genomes; $i++){
    for(my $j=0; $j < $num_genomes; $j++){
	$distance_hash{$genome_names[$i]}{$genome_names[$j]}=$matrix[$i][$j];
    }
}



# The constructor for the class
sub new {
    my ($class, %arg) = @_;
    my $self;
    #Use accession as primary identifier
    if(defined($arg{accession})){
	$self= $_genome_desc{$arg{accession}};

    }elsif(defined($arg{gi})){ #use gi identifier
	foreach(keys %_genome_desc){
	    if($_genome_desc{$_}{_gi}==$arg{gi}){
		$self= $_genome_desc{$_};
	    }
	}
    }else{
	croak "Can't recongize primary identifier. Please use gi or accesion!";
    }
    unless(defined($self)){
	croak "Can't find information for given identifier. Check for error in your script or update data source for Genome.pm";
    } 
    #add distance key
    $self->{_distance}=$distance_hash{$self->{_accession}};

    $class->_incr_count(  );  
    return bless $self, $class;
}

# This takes the place of such accessor definitions as:
#  sub get_attribute { ... }
# and of such mutator definitions as:
#  sub set_attribute { ... }
sub AUTOLOAD {
    my ($self, $newvalue) = @_;

    my ($operation, $attribute) = ($AUTOLOAD =~ /(get|set)(_\w+)$/);
    
    # Is this a legal method name?
    unless($operation && $attribute) {
        croak "Method name $AUTOLOAD is not in the recognized form (get|set)_
attribute\n";
    }
    unless(exists $self->{$attribute}) {
        croak "No such attribute '$attribute' exists in the class ", ref($self);
    }

    # Turn off strict references to enable "magic" AUTOLOAD speedup
    no strict 'refs';

    # AUTOLOAD accessors
    if($operation eq 'get') {
        # define subroutine
        *{$AUTOLOAD} = sub { shift->{$attribute} };

    # AUTOLOAD mutators
    }elsif($operation eq 'set') {
        # define subroutine
        *{$AUTOLOAD} = sub { shift->{$attribute} = shift; };

        # set the new attribute value
        $self->{$attribute} = $newvalue;
    }

    # Turn strict references back on
    use strict 'refs';

    # return the attribute value
    return $self->{$attribute};
}

# When an object is no longer being used, this will be automatically called
# and will adjust the count of existing objects
sub DESTROY {
    my($self) = @_;
    $self->_decr_count(  );
}


#Requires an accession identifier as an arguement
sub get_distance{
    shift->{_distance}{$_[0]};

}


1;

=head1 Genome

Genomes: objects for Genomes with a set of attributes

=head1 Synopsis

use Genome;

my $genome_obj = new Genome(accession => 'NC_003304');

print "Genome size is ", $genome_obj->get_size(  );

print "Genome taxonomy id is ", $genome->get_tax_id(  );

print "Genome Genus is ", $gene1->get_genus(  );

=head1 AUTHOR

Morgan Langille

=head1 Date

May 8th, 2006

=cut

1;
